/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Implementation of Asym Routing Engine.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdint.h>
#include <limits.h>
#include <unistd.h>
#include <ctype.h>
#include <complib/cl_types_osd.h>
#include <complib/cl_types.h>
#include <complib/cl_byteswap.h>
#include <complib/cl_u64_vector.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_UCAST_ASYM_C
#include <opensm/osm_switch.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_mcast_mgr.h>
#include <opensm/osm_root_detection.h>
#include <opensm/osm_max_flow_algorithm.h>
#include <opensm/osm_ucast_asym.h>
#include <opensm/osm_inform.h>
#include <opensm/osm_helper.h>

static void asym_pods_destroy(asym_t *p_asym);

/* ---------------------------------------------------------------------------------------------- */

static inline boolean_t asym_switch_is_ar_enabled(IN osm_switch_t *p_sw)
{
	return ib_general_info_is_cap_enabled(&p_sw->p_node->general_info,
					      IB_GENERAL_INFO_ADAPTIVE_ROUTING_REV_SUPPORTED_BIT);
}

static inline boolean_t asym_switch_is_isolated(IN osm_subn_t *p_subn, IN osm_switch_t *p_sw)
{
	return (p_sw->rank == OSM_SW_NO_RANK ||
		(p_sw->rank == 0 && lookup_held_back_switch(p_subn, p_sw->p_node)));
}

/* ---------------------------------------------------------------------------------------------- */

void asym_set_remote_group_ids(IN asym_t *p_asym,
			       IN osm_switch_t *p_src_sw,
			       IN osm_switch_t *p_dst_sw,
			       IN osm_ar_subgroup_t *p_group)
{
	int				i;
	uint8_t				port_type;
	uint8_t				lft_port;
	uint8_t 			new_port = 0;
	uint16_t			dst_lid;
	uint16_t			min_lid;
	uint16_t			max_lid;
	uint16_t			group_id;
	uint16_t			group_cap;
	uint16_t			group_top;
	asym_node_t			*p_anode;
	asym_node_t			*p_dst_anode;
	osm_subn_t			*p_subn;
	osm_advanced_routing_data_t	*p_ar_data;

	p_subn = p_asym->p_subn;

	p_ar_data = &p_src_sw->ar_data;
	group_top = ib_ar_info_get_group_top_ho(&p_ar_data->req_ar_info);
	group_cap = ib_ar_info_get_group_cap_ho(&p_ar_data->ar_info);


	p_anode = index_to_node(p_src_sw->re_index);
	p_dst_anode = index_to_node(p_dst_sw->re_index);

	/*
	 * Is this switch AR capable?
	 */
	if (!asym_switch_is_ar_enabled(p_src_sw)) {
		OSM_LOG(p_asym->p_log, OSM_LOG_DEBUG, "%s is not AR enabled\n", sw_name(p_src_sw));
	        return;
	}
	if (asym_switch_is_isolated(p_subn, p_src_sw) || asym_switch_is_isolated(p_subn, p_dst_sw)) {
		return;
	}

	/*
	 * Set static route to remote switch port 0.
	 */
	dst_lid = osm_node_get_base_lid(p_dst_sw->p_node, 0);
	dst_lid = cl_ntoh16(dst_lid);

	asym_set_new_static_lft_entry(p_subn, p_src_sw, dst_lid);

	/*
	 * We previously processed the ports_list and cached the lids which were attached to
	 * each switch.  We now take that list and use it to create the AR entries.
	 */
	group_id = p_dst_anode->group_id;
	if ((group_id != UCAST_MGR_INVALID_GROUP_ID) && (group_id > p_anode->max_group_id)) {
		p_anode->max_group_id = group_id;
	}

	if (group_id > group_top) {
		group_top = group_id;
	}

	for (i = 0; i < p_dst_anode->ca_count; i++) {
		port_type = p_dst_anode->ports[i].type;
		min_lid = p_dst_anode->ports[i].min_lid;
		max_lid = p_dst_anode->ports[i].max_lid;

		/*
		 * Set the AR group for these lids if needed.  (ROUTERS and AN ports get static group.)
		 */
		for (dst_lid = min_lid; dst_lid < max_lid; dst_lid++) {
			if ((p_asym->ar_state == OSM_AR_IB_LID_STATE_STATIC) ||
			    (port_type != IB_SPECIAL_PORT_TYPE_RESERVED) ||
			    (group_id == UCAST_MGR_INVALID_GROUP_ID) ||
			    (group_id >= group_cap)) {
				asym_set_new_static_lft_entry(p_subn, p_src_sw, dst_lid);
				continue;
			}

			lft_port = osm_switch_get_port_by_lid(p_src_sw, dst_lid, OSM_NEW_LFT);
			if (lft_port == OSM_NO_PATH) {
				/*
				 * Should never happen
				 */
				continue;
			}

			/*
			 * If the current static port is not in the AR group, find next port that is.
			 */
			if (!osm_ar_subgroup_check_port(p_group, lft_port)) {
				int attempts = 0;
				for (; attempts < 256; attempts++) {
					new_port++;
					if (osm_ar_subgroup_check_port(p_group, new_port)) {
						p_src_sw->new_lft[dst_lid] = new_port;
						break;
					}
				}
			}

			/*
			 * Update ARLFT
			 */
			asym_set_new_free_lft_entry(p_subn, p_src_sw, dst_lid, group_id, p_group);
		}
	}

	osm_ar_set_required_group_top(p_ar_data, group_top);
}

void asym_set_local_group_id(IN asym_t *p_asym, IN osm_switch_t *p_sw)
{
	int			i;
	uint16_t		lid;
	uint16_t		sw_lid;
	uint16_t		min_lid;
	uint16_t		max_lid;
	uint16_t		remote_lid;
	asym_node_t		*p_anode;
	osm_subn_t		*p_subn;
	osm_switch_t		*p_remote_sw;

	p_subn = p_asym->p_subn;
	p_anode = index_to_node(p_sw->re_index);

	/*
	 * Is this switch AR capable?
	 */
	if (!asym_switch_is_ar_enabled(p_sw)) {
		OSM_LOG(p_asym->p_log, OSM_LOG_DEBUG, "%s is not AR enabled\n", sw_name(p_sw));
		return;
	}

	if (asym_switch_is_isolated(p_subn, p_sw)) {
		return;
	}

	/*
	 * Set a static route to port 0.
	 */
	sw_lid = osm_node_get_base_lid(p_sw->p_node, 0);
	sw_lid = cl_ntoh16(sw_lid);
	asym_set_new_static_lft_entry(p_subn, p_sw, sw_lid);

	/*
	 * Set a static route to attached CAs.
	 */
	for (i = 0; i < p_anode->ca_count; i++) {
		min_lid = p_anode->ports[i].min_lid;
		max_lid = p_anode->ports[i].max_lid;

		for (lid = min_lid; lid < max_lid; lid++) {
			asym_set_new_static_lft_entry(p_subn, p_sw, lid);
		}
	}

	/*
	 * Set a static route to remote switch ports.
	 */
	foreach_switch(p_subn, p_remote_sw) {
		if (p_remote_sw != p_sw && !asym_switch_is_isolated(p_subn, p_remote_sw)) {
			remote_lid = osm_node_get_base_lid(p_remote_sw->p_node, 0);
			remote_lid = cl_ntoh16(remote_lid);
			asym_set_new_static_lft_entry(p_subn, p_sw, remote_lid);

			p_anode = index_to_node(p_remote_sw->re_index);
			for (i = 0; i < p_anode->ca_count; i++) {
				min_lid = p_anode->ports[i].min_lid;
				max_lid = p_anode->ports[i].max_lid;

				for (lid = min_lid; lid < max_lid; lid++) {
					asym_set_new_static_lft_entry(p_subn, p_sw, lid);
				}
			}
		}
	}
}

static int asym_set_local_group_ids(IN void **tasks,
				    IN unsigned task_len,
				    IN void *context,
				    IN unsigned int thread_id)
{
	int			i;
	asym_t			*p_asym;

	p_asym = (asym_t *)context;
	OSM_LOG_ENTER(p_asym->p_log);

	for (i = 0; i < task_len; i++) {
		asym_set_local_group_id(p_asym, (osm_switch_t *)tasks[i]);
	}

	OSM_LOG_EXIT(p_asym->p_log);
	return 0;
}

/* ---------------------------------------------------------------------------------------------- */

static asym_pod_t *asym_pod_create(IN asym_t *p_asym, IN osm_switch_t *p_sw)
{
	int			leaf_index;
	int			spine_index;
	boolean_t		found;
	asym_pod_t		*p_pod;
	asym_node_t		*p_anode;
	asym_node_t		*p_rnode;
	flow_vertex_t		*p_vertex;
	flow_vertex_t		*p_rvertex;
	flow_edge_t		*p_edge;

	/*
	 * Create the pod.
	 */
	if ((p_pod = calloc(1, sizeof(asym_pod_t))) == NULL) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B104: allocation failed for pods\n");
		return NULL;
	}

	p_asym->num_pods++;
	p_pod->index = p_asym->num_pods;

	/*
	 * Seed the pod with the switch.
	 */
	asym_set_create(&p_pod->spines);
	asym_set_create(&p_pod->leaves);

	asym_set_insert(&p_pod->leaves, p_sw);
	p_anode = index_to_node(p_sw->re_index);
	p_anode->pod = p_pod->index;


	leaf_index = 1;
	spine_index = 1;

	/*
	 * Map all of the leaves which share spines with this switch.
	 */
	do {
		found = FALSE;

		/*
		 * Find the spines which are connected to the new leaves we
		 * found on the previous iteration.
		 */
		leaf_index--;
		while (leaf_index++, leaf_index < asym_set_size(&p_pod->leaves)) {
			p_sw = (osm_switch_t *)asym_set_get(&p_pod->leaves, leaf_index);
			p_vertex = index_to_vertex(p_sw->re_index);
			p_anode = index_to_node(p_sw->re_index);

			foreach_vertex_edge(p_asym->p_graph, p_vertex, p_edge) {
				p_rvertex = index_to_vertex(p_edge->adj_index);
				p_rnode = index_to_node(p_edge->adj_index);

				if (asym_switch_is_spine(p_rvertex->p_sw)) {
					p_anode->num_connections++;

					if (p_rnode->pod == 0) {
						asym_set_insert(&p_pod->spines, p_rvertex->p_sw);
						p_rnode->pod = p_pod->index;
						found = TRUE;
					}
				}
			}
		}

		/*
		 * Find the leaves which are connected to the new spines we
		 * found on the previous iteration.
		 */
		spine_index--;
		while (spine_index++, spine_index < asym_set_size(&p_pod->spines)) {
			p_sw = (osm_switch_t *)asym_set_get(&p_pod->spines, spine_index);
			p_vertex = index_to_vertex(p_sw->re_index);
			p_anode = index_to_node(p_sw->re_index);

			foreach_vertex_edge(p_asym->p_graph, p_vertex, p_edge) {
				p_rvertex = index_to_vertex(p_edge->adj_index);
				p_rnode = index_to_node(p_edge->adj_index);

				if (asym_switch_is_leaf(p_rvertex->p_sw)) {
					p_anode->num_connections++;

					if (p_rnode->pod == 0) {
						asym_set_insert(&p_pod->leaves, p_rvertex->p_sw);
						p_rnode->pod = p_pod->index;
						found = TRUE;
					}
				}
			}
		}
	} while (found);

	/*
	 * Find the fully connected leaves.
	 */
	for (leaf_index = 1; leaf_index < asym_set_size(&p_pod->leaves); leaf_index++) {
		p_sw = (osm_switch_t *)asym_set_get(&p_pod->leaves, leaf_index);
		p_anode = index_to_node(p_sw->re_index);
		p_anode->is_fully_connected = (p_anode->num_connections == asym_set_size(&p_pod->spines)-1);
	}

	return p_pod;
}

static void asym_pods_create(IN asym_t *p_asym)
{
	osm_switch_t		*p_sw;
	flow_vertex_t		*p_vertex;
	asym_node_t		*p_anode;
	asym_pod_t		*p_pod;

	cl_qmap_init(&p_asym->pod_map);

	foreach_vertex(p_asym->p_graph, p_vertex) {
		p_sw = p_vertex->p_sw;
		p_anode = index_to_node(p_sw->re_index);
		if ((p_sw->rank == ASYM_RANK_LEAF) && (p_anode->pod == 0)) {
			if ((p_pod = asym_pod_create(p_asym, p_sw)) != NULL) {
				cl_qmap_insert(&p_asym->pod_map, p_pod->index, (cl_map_item_t *)p_pod);
			}
		}
	}
}

static void asym_pods_destroy(IN asym_t *p_asym)
{
	cl_map_item_t		*p_item;
	cl_qmap_t		*p_map;
	asym_pod_t		*p_pod;

	p_map = &p_asym->pod_map;
	while ((p_item = cl_qmap_head(p_map)) != cl_qmap_end(p_map)) {
		p_pod = (asym_pod_t *)p_item;

		asym_set_remove_all(&p_pod->spines);
		asym_set_destroy(&p_pod->spines);

		asym_set_remove_all(&p_pod->leaves);
		asym_set_destroy(&p_pod->leaves);

		cl_qmap_remove(p_map, p_pod->index);
		free(p_pod);
	}

	cl_qmap_remove_all(p_map);
}

/* ---------------------------------------------------------------------------------------------- */

static int asym_validate_tree(osm_subn_t *p_subn, flow_graph_t *p_graph)
{
	int			max_rank;
	int			port_type;
	int			num_leaves;
	int			min_num_groups;
	osm_port_t		*p_port;
	osm_physp_t		*p_physp;
	osm_switch_t		*p_sw;
	flow_vertex_t		*p_vertex;
	flow_edge_t		*p_edge1;
	flow_edge_t		*p_edge2;

	/*
	 * A switch with rank > 2 is only an error if it has attached HCAs.  We check this by
	 * finding all switches connected to HCAs and verifying their ranks.
	 */
	foreach_port(p_subn, p_port) {
		p_physp = p_port->p_physp;

		if (p_port->p_node->sw != NULL) {
			continue;
		} else if (p_physp == NULL) {
			continue;
		} else if (p_physp->p_remote_physp == NULL) {
			continue;
		}

		port_type = osm_phsyp_get_special_port_type(p_physp);
		if (port_type == IB_SPECIAL_PORT_TYPE_RESERVED) {
			p_sw = p_port->p_remote_sw_node->sw;
			if (p_sw->rank > ASYM_RANK_LEAF) {
				OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO, "leaf switch with rank %d detected\n", p_sw->rank);
				return -1;
			}
		}

	}

	/*
	 * Find the maximum rank.  It must not be greater than 2.  We previously checked for leaf switches
	 * with rank > 2.  If we encounter a switch with rank > 2 now, it is NOT a leaf switch.
	 *
	 * In addition, count number of leaves to ensure we have enough AR group IDs to route the subnet.
	 */
	max_rank = 0;
	num_leaves = 0;
	min_num_groups = 0xffff;
	foreach_switch(p_subn, p_sw) {
		if (!lookup_held_back_switch(p_subn, p_sw->p_node)) {
			if (p_sw->rank >= ASYM_RANK_LEAF) {
				max_rank = ASYM_RANK_LEAF;
			}
			if (p_sw->rank == ASYM_RANK_LEAF) {
				num_leaves++;
			}
			if (p_sw->ar_data.num_groups < min_num_groups) { ///DK REQ
				min_num_groups = p_sw->ar_data.num_groups;
			}
		}
	}

	if (max_rank != ASYM_RANK_LEAF) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO, "tree is not 3 levels (max_rank = %d)\n", max_rank);
		return -1;
	}

	if (min_num_groups < num_leaves) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO, "number of leaves exceeds switch capabilites"
			" (num leaves = %d, minimal group cap = %d)\n",
			num_leaves, min_num_groups);
		return -1;
	}

	/*
	 * Search for parallel links.
	 */
	foreach_vertex(p_graph, p_vertex) {
		foreach_vertex_edge(p_graph, p_vertex, p_edge1) {
			foreach_vertex_edge(p_graph, p_vertex, p_edge2) {
				if ((p_edge1 != p_edge2) && (p_edge1->adj_index == p_edge2->adj_index)) {
					OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO, "parallel links detected\n");
					return -1;
				}
			}
		}
	}

	OSM_LOG(&p_subn->p_osm->log, OSM_LOG_VERBOSE, "tree is asym compliant\n");
	return 0;
}

static void asym_execute_threaded(IN asym_t *p_asym, IN void *function, IN int task_type)
{
	int			count;
	size_t			task_len;
	asym_pod_t		*p_pod;
	osm_subn_t		*p_subn;
	osm_ucast_mgr_t		*p_mgr;
	osm_switch_t		*p_sw;
	void			**tasks;

	p_subn = p_asym->p_subn;
	p_mgr = &p_subn->p_osm->sm.ucast_mgr;

	/*
	 * Allocate and load the task list.
	 */
	if (task_type == ASYM_TASK_SWITCHES) {
		task_len = p_asym->num_switches;
	} else if (task_type == ASYM_TASK_PODS) {
		task_len = p_asym->num_pods;
	} else {
		return;
	}

	if ((tasks = (void **)calloc(task_len, sizeof(void *))) == NULL) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B106: allocation failed for tasks\n");
		return;
	}

	if (task_type == ASYM_TASK_SWITCHES) {
		count = 0;
		foreach_switch(p_subn, p_sw) {
			if (!lookup_held_back_switch(p_subn, p_sw->p_node)) {
				tasks[count++] = (void *)p_sw;
			}
		}
	} else if (task_type == ASYM_TASK_PODS) {
		count = 0;
		foreach_asym_pod(p_asym, p_pod) {
			tasks[count++] = (void *)p_pod;
		}
	}

	/*
	 * Execute the function.
	 */
	cl_mp_thread_pool_array_apply(&p_mgr->thread_pool,
				      tasks,
				      count,
				      CL_MP_THREAD_POOL_BATCH_PER_THREAD,
				      (cl_pfn_mp_thread_pool_apply_t)function,
				      (void *)p_asym);

	free(tasks);
}

/* ---------------------------------------------------------------------------------------------- */

void asym_set_new_lft_entry(IN osm_subn_t *p_subn,
			    IN osm_switch_t *p_sw,
			    IN uint16_t lid,
			    IN int state,
			    IN int id)
{
	uint8_t			lft_port;

	lft_port = osm_switch_get_port_by_lid(p_sw, lid, OSM_NEW_LFT);
	if (lft_port != OSM_NO_PATH) {
		osm_ar_lft_set(&p_sw->ar_data, lid, 0, lft_port, state, id);
	} else {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_VERBOSE, "ERR B105: %s - no port found for lid %d\n",
			p_sw->p_node->print_desc, lid);
	}
}

void asym_set_new_static_lft_entry(IN osm_subn_t *p_subn,
				   IN osm_switch_t *p_sw,
				   IN uint16_t lid)
{
	asym_set_new_lft_entry(p_subn, p_sw, lid, OSM_AR_IB_LID_STATE_STATIC, 0);
}

void asym_set_new_free_lft_entry(IN osm_subn_t *p_subn,
				     IN osm_switch_t *p_sw,
				     IN uint16_t lid,
				     IN uint16_t group_id,
				     IN osm_ar_subgroup_t *ar_group)
{
	osm_ar_subgroup_t		*p_subgroup;
	osm_advanced_routing_data_t	*p_ar_data;

	p_ar_data = &p_sw->ar_data;

	/*
	 * Set the subgroup with the remote switch's AR group.
	 */
	p_subgroup = osm_ar_subgroup_get(p_ar_data, group_id, 0);
	memcpy(p_subgroup, ar_group, sizeof(osm_ar_subgroup_t));
	asym_set_new_lft_entry(p_subn, p_sw, lid, OSM_AR_IB_LID_STATE_FREE, group_id);
}

/* ---------------------------------------------------------------------------------------------- */

static void asym_ar_resolve_pod(IN asym_t *p_asym, IN asym_pod_t *p_pod, IN void *context)
{
	asym_pod_t			*p_rpod;
	osm_subn_t			*p_subn;

	p_subn = p_asym->p_subn;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	foreach_asym_pod(p_asym, p_rpod) {
		asym_algorithm_pod_to_pod(context, p_pod, p_rpod);
	}

	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return;
}

static int asym_ar_resolve_pods(IN void **tasks,
				IN unsigned task_len,
				IN void *context,
				IN unsigned int thread_id)
{
	int			i;
	void			*p_algo;
	asym_t			*p_asym;

	p_asym = (asym_t *)context;
	OSM_LOG_ENTER(&p_asym->p_subn->p_osm->log);

	if ((p_algo = asym_algorithm_init(p_asym)) != NULL) {
		for (i = 0; i < task_len; i++) {
			asym_ar_resolve_pod(p_asym, tasks[i], p_algo);
		}


		asym_algorithm_free(p_algo);
	}

	OSM_LOG_EXIT(&p_asym->p_subn->p_osm->log);
	return 0;
}

/* ========================================================================================================
 *
 * API functions.
 */
int osm_asym_ar_process_switch(IN osm_subn_t *p_subn,
			       IN osm_switch_t *p_sw,
			       IN void *context)
{
	asym_t			*p_asym;
	asym_node_t		*p_anode;

	p_asym = (asym_t *)context;
	p_anode = index_to_node(p_sw->re_index);
	return p_anode->max_group_id + 1;
}

int osm_asym_ar_process_subnet(IN osm_subn_t *p_subn,
				IN int ar_enable,
				IN int ar_mode,
				IN void *context)
{
	int			port_type;
	uint16_t		sw_lid;
	uint16_t		min_lid;
	uint16_t		max_lid;
	asym_t			*p_asym;
	asym_node_t		*p_anode;
	asym_node_t		*p_rnode;
	osm_port_t		*p_port;
	osm_node_t		*p_remote_node;
	osm_physp_t		*p_physp;
	osm_switch_t		*p_remote_sw;
	struct timeval		start, stop;
	double			elapsed_time;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	p_asym = (asym_t *)context;
	p_asym->ar_enable = ar_enable;
	p_asym->ar_mode = ar_mode;
	p_asym->ar_state = ar_enable ? ar_mode : OSM_AR_IB_LID_STATE_STATIC;

	/*
	 * Find the leaf switches and the lids of the attached HCAs.
	 */
	foreach_asym_node(p_asym, p_anode) {
		p_anode->ca_count = 0;
	}

	foreach_port(p_asym->p_subn, p_port) {
		p_physp = p_port->p_physp;

		if (p_port->p_node->sw != NULL) {
			continue;
		} else if (p_physp == NULL) {
			continue;
		} else if (p_physp->p_remote_physp == NULL) {
			continue;
		}

		port_type = osm_phsyp_get_special_port_type(p_physp);

		/*
		 * The port in question is either a CA port or a special port.
		 * CA ports get an AR group, while special ports do not.
		 */
		osm_port_get_lid_range_ho(p_port, &min_lid, &max_lid);
		max_lid++;
		if (!min_lid) {
			continue;
		}

		p_remote_node = p_port->p_remote_sw_node;
		p_remote_sw = p_remote_node->sw;

		p_rnode = index_to_node(p_remote_sw->re_index);
		p_rnode->ports[p_rnode->ca_count].type = port_type;
		p_rnode->ports[p_rnode->ca_count].min_lid = min_lid;
		p_rnode->ports[p_rnode->ca_count].max_lid = max_lid;
		p_rnode->ca_count++;

		if ((p_remote_sw->rank == ASYM_RANK_LEAF) && (port_type == IB_SPECIAL_PORT_TYPE_RESERVED)) {
			p_rnode->is_leaf_switch = TRUE;
		}
	}

	/*
	 * Set the group ids.  This is done here because locking isn't allowed in the routing code.
	 * If we tried to do it later, it might be in the parallel threading model.
	 */
	foreach_asym_node(p_asym, p_anode) {
		if (p_anode->is_leaf_switch) {
			sw_lid = osm_node_get_base_lid(p_anode->p_sw->p_node, 0);
			sw_lid = cl_ntoh16(sw_lid);
			p_anode->group_id = osm_ucast_mgr_make_groupid(&p_subn->p_osm->sm.ucast_mgr, sw_lid, 0);
		}
	}

	gettimeofday(&start, NULL);

	/*
	 * Set the local AR groups.
	 */
	asym_execute_threaded(p_asym, asym_set_local_group_ids, ASYM_TASK_SWITCHES);

	/*
	 * Run the algorithm and set the remote AR groups.
	 */
	asym_execute_threaded(p_asym, asym_ar_resolve_pods, ASYM_TASK_PODS);

	gettimeofday(&stop, NULL);
	elapsed_time = (double)(stop.tv_usec - start.tv_usec) / 1000000 + (double)(stop.tv_sec - start.tv_sec);
	OSM_LOG(p_asym->p_log, OSM_LOG_DEBUG, "time to solution = %11.8f secs\n", elapsed_time);

	OSM_LOG_EXIT(&p_subn->p_osm->log);

	return 0;
}

void osm_asym_free(IN void *context)
{
	asym_node_t		*p_anode;
	asym_t			*p_asym;

	p_asym = (asym_t *)context;

	if (p_asym != NULL) {
		asym_pods_destroy(p_asym);

		foreach_anode(p_asym, p_anode) {
			free(p_anode->links);
		}

		free(p_asym->nodes);
		free(p_asym->links);

		if (p_asym->p_graph != NULL) {
			osm_flow_destroy_graph(&p_asym->p_graph);
		}

		memset(p_asym, 0, sizeof(asym_t));
		free(p_asym);
	}
}

void *osm_asym_init(IN osm_subn_t *p_subn)
{
	int			status;
	int			num_links;
	int			num_edges;
	size_t			num_switches;
	flow_edge_t		*p_edge;
	flow_graph_t		*p_graph;
	flow_vertex_t		*p_vertex;
	asym_node_t		*p_anode;
	asym_node_t		*p_rnode;
	asym_t			*p_asym;
	osm_switch_t		*p_sw;

	OSM_LOG_ENTER(&p_subn->p_osm->log);

	/*
	 * Validate the tree.  It must be 3-level without parallel links.
	 */
	p_graph = osm_flow_build_graph(p_subn);
	if (asym_validate_tree(p_subn, p_graph) < 0) {
		osm_flow_destroy_graph(&p_graph);
		return NULL;
	}

	/*
	 * The tree is valid.
	 */
	if ((p_asym = (asym_t *)calloc(1, sizeof(asym_t))) == NULL) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR B101: allocation failed for asym\n");
		osm_flow_destroy_graph(&p_graph);
		return NULL;
	}

	/*
	 * Repopulate the data structs.
	 */
	num_switches = p_graph->vertices_num;
	num_edges = p_graph->num_edges;

	p_asym->p_osm = p_subn->p_osm;
	p_asym->p_subn = p_subn;
	p_asym->p_log = &p_subn->p_osm->log;
	p_asym->p_mgr = &p_subn->p_osm->sm.ucast_mgr;
	p_asym->p_graph = p_graph;
	p_asym->num_switches = num_switches;

	p_asym->nodes = (asym_node_t *)calloc(num_switches, sizeof(asym_node_t));
	p_asym->links = (asym_link_t *)calloc(num_edges, sizeof(asym_link_t));

	if ((p_asym->nodes == NULL) || (p_asym->links == NULL)) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B102: allocation failed for asym\n");
		status = -1;
		goto error_exit;
	}

	/*
	 * Label the switches with their indices.
	 */
	foreach_vertex(p_asym->p_graph, p_vertex) {
		num_links = p_vertex->end_edges_index - p_vertex->start_edges_index;

		p_sw = p_vertex->p_sw;
		p_sw->re_index = p_vertex->index;
		p_sw->re_priv = p_asym;

		p_anode = index_to_node(p_vertex->index);
		p_anode->p_sw = p_sw;
		p_anode->lid = cl_ntoh16(osm_node_get_base_lid(p_sw->p_node, 0));
		p_anode->index = p_sw->re_index;

		if ((p_anode->links = (asym_link_t **)calloc(num_links+1, sizeof(asym_link_t *))) == NULL) {
			OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B103: allocation failed for asym node data\n");
			status = -1;
			goto error_exit;
		}
	}

	/*
	 * Find the ports between switches.
	 */
	foreach_edge(p_asym->p_graph, p_edge) {
		p_asym->links[p_asym->num_links].index  = p_edge - p_asym->p_graph->edges;
		p_asym->links[p_asym->num_links].index1 = p_edge->index;
		p_asym->links[p_asym->num_links].port1  = p_edge->port_num;
		p_asym->links[p_asym->num_links].index2 = p_edge->adj_index;
		p_asym->links[p_asym->num_links].port2  = p_edge->adj_port_num;

		if (p_edge->index < p_edge->adj_index) {
			p_anode = index_to_node(p_edge->index);
			p_anode->links[p_anode->num_links++] = &p_asym->links[p_asym->num_links];

			p_rnode = index_to_node(p_edge->adj_index);
			p_rnode->links[p_rnode->num_links++] = &p_asym->links[p_asym->num_links];
		}

		p_asym->num_links++;
	}

	/*
	 *  Create the pods.
	 */
	asym_pods_create(p_asym);

	status = 0;

error_exit:
	if (status < 0) {
		osm_asym_free(p_asym);
		p_asym = NULL;
	}

	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return p_asym;
}
