/*
 * Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Implementation of Asym Algorithm.
 */

#include <math.h>
#include <regex.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <complib/cl_types.h>
#include <complib/cl_byteswap.h>
#include <complib/cl_u64_vector.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_UCAST_ASYM_ALGORITHM_C
#include <opensm/osm_max_flow_algorithm.h>
#include <opensm/osm_ucast_asym.h>

#define	asym_min(A,B)		(((A) < (B)) ? (A) : (B))

/*
 * Debug defines only used in the standalone test program.  NOT to be used by SM.
 */
#ifndef	ASYM_DEBUG
#define	layer_print_all(HEADER, P_ALGO, P_LAYER)
#endif


/* ========================================================================================================
 *
 * Utility functions.
 */

static asym_adata_t *node_choose(IN asym_algorithm_t *p_algo, IN asym_adata_t *p_adata, IN asym_adata_t *p_bdata)
{
	asym_t		*p_asym = p_algo->p_asym;
	osm_switch_t	*p_asw = adata_to_switch(p_adata);
	osm_switch_t	*p_bsw = adata_to_switch(p_bdata);

	return (strcmp(sw_name(p_asw), sw_name(p_bsw)) <= 0) ? p_adata : p_bdata;
}

static asym_alink_t *link_choose(IN asym_algorithm_t *p_algo, IN asym_adata_t *p_adata, IN asym_alink_t *p_alink, IN asym_alink_t *p_blink)
{
	asym_t		*p_asym = p_algo->p_asym;
	osm_switch_t	*p_asw = alink_remote_switch(p_adata, p_alink);
	osm_switch_t	*p_bsw = alink_remote_switch(p_adata, p_blink);

	return (strcmp(sw_name(p_asw), sw_name(p_bsw)) <= 0) ? p_alink : p_blink;
}

static asym_alink_t *link_find(IN asym_algorithm_t *p_algo, IN osm_switch_t *p_sw1, IN osm_switch_t *p_sw2)
{
	osm_switch_t            *p_a_sw1;
	osm_switch_t            *p_a_sw2;
	asym_alink_t            *p_alink;
	asym_adata_t            *p_adata;
	asym_iter_t             iter;
	asym_t			*p_asym;

	p_asym = p_algo->p_asym;

	p_adata = switch_to_adata(p_sw1);
	foreach_link(p_adata, p_alink, iter) {
		alink_to_switches(p_alink, p_a_sw1, p_a_sw2);

		if ((p_a_sw1 == p_sw1) && (p_a_sw2 == p_sw2)) {
			return p_alink;
		}

		if ((p_a_sw2 == p_sw1) && (p_a_sw1 == p_sw2)) {
			return p_alink;
		}
	}

	return NULL;
}

/* ========================================================================================================
 *
 * Infrastructure functions.
 */

void asym_algorithm_reset(IN void *context)
{
	int			i;
	osm_switch_t		*p_sw;
	asym_algorithm_t	*p_algo;
	asym_adata_t		*p_adata;

	p_algo = (asym_algorithm_t *)context;

	if (p_algo != NULL) {
		for (i = 0; i < MAX_LAYERS; i++) {
			asym_set_remove_all(&p_algo->layers[i]);
		}

		asym_set_remove_all(&p_algo->removals);

		p_algo->p_src_sw = NULL;
		p_algo->p_dst_sw = NULL;

		p_algo->layer_number = 0;
		p_algo->retries = 0;

		foreach_switch(p_algo->p_asym->p_subn, p_sw) {
			p_adata = switch_to_adata(p_sw);
			p_adata->layer = 0;
		}
	}
}

void asym_algorithm_free(IN void *context)
{
	int			i;
	asym_algorithm_t	*p_algo;

	p_algo = (asym_algorithm_t *)context;

	if (p_algo != NULL) {
		for (i = 0; i < MAX_LAYERS; i++) {
			asym_set_remove_all(&p_algo->layers[i]);
			asym_set_destroy(&p_algo->layers[i]);
		}

		asym_set_remove_all(&p_algo->removals);
		asym_set_destroy(&p_algo->removals);

		free(p_algo->p_alink);
		free(p_algo->p_adata);
		free(p_algo->distances);
		free(p_algo->visited);
		free(p_algo->queue);

		memset(p_algo, 0, sizeof(asym_algorithm_t));
		free(p_algo);
	}
}

void *asym_algorithm_init(IN asym_t *p_asym)
{
	int			i;
	int			status = 0;
	asym_node_t		*p_anode = NULL;
	asym_alink_t		*p_alink = NULL;
	asym_adata_t		*p_adata = NULL;
	asym_adata_t		*p_rdata = NULL;
	asym_algorithm_t	*p_algo = NULL;
	flow_edge_t		*p_edge = NULL;

	/*
	 * Allocate the main data structure.
	 */
	if ((p_algo = calloc(1, sizeof(asym_algorithm_t))) == NULL) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B201 - allocation failure for p_algo\n");
		status = -1;
		goto exit;
	}

	/*
	 * Create the removals set for setting up AR groups.
	 */
	asym_set_create(&p_algo->removals);

	/*
	 * Create the layers.
	 */
	for (i = 0; i < MAX_LAYERS; i++) {
		asym_set_create(&p_algo->layers[i]);
	}

	/*
	 * BFS data structs.
	 */
	p_algo->distances_size = (p_asym->num_switches + 1) *sizeof(uint8_t);
	p_algo->visited_size   = (p_asym->num_switches + 1) *sizeof(uint8_t);
	p_algo->queue_size     = (p_asym->num_switches + 1) *sizeof(uint16_t);

	p_algo->distances = calloc(1, p_algo->distances_size);
	p_algo->visited   = calloc(1, p_algo->visited_size);
	p_algo->queue     = calloc(1, p_algo->queue_size);

	if ((p_algo->distances == NULL) || (p_algo->visited == NULL) || (p_algo->queue == NULL)) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B202 - allocation failure for BFS data\n");
		status = -1;
		goto exit;
	}

	/*
	 * Algorithm structs.
	 */
	p_algo->alink_size = (p_asym->num_links + 1) * sizeof(asym_alink_t);
	p_algo->adata_size = (p_asym->num_switches + 1) * sizeof(asym_adata_t);

	p_algo->p_alink = calloc(1, p_algo->alink_size);
	p_algo->p_adata = calloc(1, p_algo->adata_size);

	if ((p_algo->p_alink == NULL) || (p_algo->p_adata == NULL)) {
		OSM_LOG(p_asym->p_log, OSM_LOG_ERROR, "ERR B203 - allocation failure for algorithm data\n");
		status = -1;
		goto exit;
	}

	p_algo->p_asym = p_asym;

	/*
	 * Label the data struct indices.
	 */
	foreach_anode(p_asym, p_anode) {
		p_adata = index_to_adata(p_anode->index);
		p_adata->index = p_anode->index;

		p_algo->num_adata++;
	}

	/*
	 * Label the link struct edges.
	 */
	p_alink = p_algo->p_alink;
	foreach_edge(p_asym->p_graph, p_edge) {
		if (p_edge->index < p_edge->adj_index) {
			p_alink->p_edge = p_edge;
			p_alink->index = p_algo->num_alink;
			p_alink++;

			p_algo->num_alink++;
		}
	}

	/*
	 * Map the data -> links.
	 */
	foreach_alink(p_algo, p_alink) {
		p_edge = p_alink->p_edge;

		p_adata = index_to_adata(p_edge->index);
		p_adata->p_links[p_adata->num_links++] = p_alink;

		p_rdata = index_to_adata(p_edge->adj_index);
		p_rdata->p_links[p_rdata->num_links++] = p_alink;
	}

exit:
	/*
	 * Cleanup if necessary.
	 */
	if (status < 0) {
		asym_algorithm_free(p_algo);
		p_algo = NULL;
	}

	return p_algo;
}

/* -------------------------------------------------------------------------------------------------------- */

__attribute__((unused))
static void asym_algorithm_set_ar_groups(IN asym_algorithm_t *p_algo, IN int start_layer, IN int end_layer)
{
	int			layer_number;
	uint8_t			port_num;
	uint16_t		dst_lid;
	asym_t			*p_asym;
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	asym_alink_t		*p_alink;
	asym_adata_t		*p_adata;
	osm_switch_t		*p_sw;
	osm_switch_t		*p_dst_sw;
	osm_ar_subgroup_t	dst_ar_group;

	p_asym = p_algo->p_asym;

	p_dst_sw = p_algo->p_dst_sw;
	dst_lid = osm_node_get_base_lid(p_dst_sw->p_node, 0);
	dst_lid = cl_ntoh16(dst_lid);

	/*
	 * Create the AR group for each switch in the layers.
	 */
	for (layer_number = start_layer; layer_number <= end_layer; layer_number++) {
		foreach_layer_switch(&p_algo->layers[layer_number], p_sw, iter1) {
			osm_switch_clear_and_get_minhop_ports(p_sw, dst_lid, &dst_ar_group, NULL);

			/*
			 * The CORE nodes and the DST SPINE nodes will NOT have any removals.
			 * Hence, the minhop ports are correct.  There will not be anything
			 * to remove.
			 */
			if ((layer_number != LAYER_DST_SPINE) && (layer_number != LAYER_CORE)) {
				p_adata = switch_to_adata(p_sw);
				foreach_link_removal(&p_algo->removals, p_alink, iter2) {
					if (alink_is_connected_to_adata(p_adata, p_alink) && link_is_incoming(p_adata, p_alink)) {
						OSM_LOG(p_algo->p_asym->p_log, OSM_LOG_DEBUG, "REMOVAL for (%s,%s) : (%s,%s)\n",
							sw_name(p_algo->p_src_sw), sw_name(p_algo->p_dst_sw),
							sw_name(alink_local_switch(p_adata, p_alink)), sw_name(alink_remote_switch(p_adata, p_alink)));

						port_num = alink_local_port_num(p_adata, p_alink);
						osm_ar_subgroup_sub_port(&dst_ar_group, port_num);
					}
				}
			}

			asym_set_remote_group_ids(p_algo->p_asym, p_sw, p_dst_sw, &dst_ar_group);
		}
	}
}

/* -------------------------------------------------------------------------------------------------------- */

__attribute__((unused))
static boolean_t asym_node_is_connected(IN asym_algorithm_t *p_algo, IN asym_adata_t *p_adata)
{
	asym_t			*p_asym;
	asym_iter_t		iter;
	asym_alink_t		*p_alink;
	osm_switch_t		*p_sw;

	p_asym = p_algo->p_asym;

	/*
	 * If we have at least one outgoing link which is UP
	 *	if the node is in the CORE layer, return TRUE
	 *	if the node is in the SRC SPINE layer, the connection must be to the SRC
	 */
	foreach_outgoing_link(p_adata, p_alink, iter) {
		if (p_adata->layer != LAYER_SRC_SPINE) {
			return TRUE;
		}

		p_sw = alink_remote_switch(p_adata, p_alink);
		if (p_sw == p_algo->p_src_sw) {
			return TRUE;
		}
	}

	return FALSE;
}

__attribute__((unused))
static boolean_t asym_node_is_not_connected(IN asym_algorithm_t *p_algo, IN asym_adata_t *p_adata)
{
	return !asym_node_is_connected(p_algo, p_adata);
}

__attribute__((unused))
static osm_switch_t *asym_find_fully_connected_switch(IN asym_algorithm_t *p_algo, IN asym_pod_t *p_pod)
{
	osm_switch_t		*p_sw;
	asym_iter_t		iter;
	asym_node_t		*p_anode;
	asym_t			*p_asym;

	p_asym = p_algo->p_asym;

	foreach_pod_leaf(p_pod, p_sw, iter) {
		p_anode = switch_to_anode(p_sw);
		if (p_anode->is_fully_connected) {
			return p_sw;
		}
	}

	return NULL;
}

__attribute__((unused))
static osm_switch_t *asym_find_partially_connected_switch(IN asym_algorithm_t *p_algo, IN asym_pod_t *p_pod)
{
	osm_switch_t		*p_sw;
	asym_iter_t		iter;
	asym_node_t		*p_anode;
	asym_t			*p_asym;

	p_asym = p_algo->p_asym;

	foreach_pod_leaf(p_pod, p_sw, iter) {
		p_anode = switch_to_anode(p_sw);
		if (!p_anode->is_fully_connected) {
			return p_sw;
		}
	}

	return NULL;
}

/* -------------------------------------------------------------------------------------------------------- */

__attribute__((unused))
static void asym_ar_bfs(asym_algorithm_t *p_algo)
{
	int		i;
	int		v;
	int		l_adj;
	int		q_max;
	int		hop_count;
	int		dst_index;
	uint8_t		*distances;
	uint8_t		*visited;
	uint8_t		hop1;
	uint8_t		hop2;
	uint16_t	src_lid;
	uint16_t	dst_lid;
	uint16_t	remote_lid;
	uint16_t	*queue;
	osm_switch_t	*p_src_sw;
	osm_switch_t	*p_dst_sw;
	osm_switch_t	*p_remote_sw;
	flow_vertex_t	*p_vertex;
	flow_edge_t	*p_edge;
	asym_t		*p_asym;

	p_src_sw = p_algo->p_src_sw;
	p_dst_sw = p_algo->p_dst_sw;

        src_lid = osm_node_get_base_lid(p_src_sw->p_node, 0);
        src_lid = cl_ntoh16(src_lid);

	dst_index = p_dst_sw->re_index;
        dst_lid = osm_node_get_base_lid(p_dst_sw->p_node, 0);
        dst_lid = cl_ntoh16(dst_lid);

	p_asym = p_algo->p_asym;
	distances = p_algo->distances;
	visited = p_algo->visited;
	queue = p_algo->queue;

	memset(distances, 0x00, p_algo->distances_size);
	memset(visited, 0x00, p_algo->visited_size);
	memset(queue, 0x00, p_algo->queue_size);

	q_max = 0;

	distances[dst_index] = 0;
	visited[dst_index] = 1;
	queue[q_max++] = dst_index;

	for (i = 0; i < q_max; i++) {
		v = queue[i];
		hop_count = distances[v];
		p_vertex = &p_asym->p_graph->vertices[v];

		foreach_vertex_edge(p_asym->p_graph, p_vertex, p_edge) {
			l_adj = p_edge->adj_index;

			if (visited[l_adj] != 0) {
				continue;
			}

			p_remote_sw = p_asym->p_graph->vertices[l_adj].p_sw;
			remote_lid = osm_node_get_base_lid(p_remote_sw->p_node, 0);
			remote_lid = cl_ntoh16(remote_lid);

			if (p_remote_sw->rank != abs(1-hop_count)) {
				continue;
			}

			if ((hop1 = osm_switch_get_least_hops(p_remote_sw, src_lid)) == OSM_NO_PATH) {
				continue;
			}

			if ((hop2 = osm_switch_get_least_hops(p_dst_sw, remote_lid)) == OSM_NO_PATH) {
				continue;
			}

			if ((hop1+hop2) != 4) {
				continue;
			}

			/*
			 * Add this node to the queue.
			 */
			visited[l_adj] = 1;
			distances[l_adj] = hop_count + 1;
			queue[q_max++] = l_adj;
		}
	}

	p_algo->q_max = q_max;
}

/* -------------------------------------------------------------------------------------------------------- */

static inline void asym_layer_initialize(asym_algorithm_t *p_algo, asym_set_t *p_layer)
{
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	asym_adata_t		*p_adata;
	asym_alink_t		*p_alink;
	osm_switch_t		*p_sw;

	foreach_layer_switch(p_layer, p_sw, iter1) {
		p_adata = switch_to_adata(p_sw);
		p_adata->nw = 0.0;
		p_adata->balanced = FALSE;

		foreach_link(p_adata, p_alink, iter2) {
			p_alink->state = LINK_UP;
			p_alink->rlw = 0.0;
			p_alink->lw = 0.0;
		}
	}
}

/* -------------------------------------------------------------------------------------------------------- */

__attribute__((unused))
static void asym_algorithm_setup(asym_algorithm_t *p_algo, IN osm_switch_t *p_src, IN osm_switch_t *p_dst)
{
	int			i;
	uint8_t			d;
	uint8_t			*distances;
	uint16_t		q;
	uint16_t		*queue;
	asym_adata_t		*p_adata;
	asym_adata_t		*p_rdata;
	osm_switch_t		*p_sw;
	asym_t			*p_asym;

	p_asym = p_algo->p_asym;

	/*
	 * Clean up the previous run.
	 */
	asym_algorithm_reset(p_algo);

	/*
	 * Cache the endpoints.
	 */
        p_algo->p_src_sw = p_src;
        p_algo->p_dst_sw = p_dst;

	/*
	 * Get the minhop table.
	 */
        asym_ar_bfs(p_algo);

	p_adata = index_to_adata(p_algo->p_dst_sw->re_index);

	distances = p_algo->distances;
	queue     = p_algo->queue;

	/*
	 * Create the layers for the algorithm.
	 */
	p_adata->layer = 1;

	for (i = 0; q = queue[i], d = distances[q], i < p_algo->q_max; i++) {
		d = distances[q];
		p_sw = index_to_switch(q);

		asym_set_insert(&p_algo->layers[d+1], p_sw);
		p_rdata = index_to_adata(q);
		p_rdata->layer = d+1;
	}

	/*
	 * Clear the removals.
	 */
	asym_set_remove_all(&p_algo->removals);

	/*
	 * Initialize the layers.
	 */
	for (i = LAYER_DST; i <= LAYER_SRC; i++) {
		asym_layer_initialize(p_algo, &p_algo->layers[i]);
	}

	return;
}

__attribute__((unused))
static void asym_algorithm_reset_links(asym_algorithm_t *p_algo, int layer)
{
	int			i;
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	asym_adata_t		*p_adata;
	asym_alink_t		*p_alink;
	osm_switch_t		*p_sw;

	/*
	 * Clear the fake removals.
	 */
	for (i = LAYER_DST_SPINE; i < layer; i++) {
		foreach_layer_switch(&p_algo->layers[i], p_sw, iter1) {
			p_adata = switch_to_adata(p_sw);
			foreach_link(p_adata, p_alink, iter2) {
				if (link_is_outgoing(p_adata, p_alink)) {
					if (p_alink->state != LINK_NA) {
						p_alink->state = LINK_UP;
					}
				}
			}
		}
	}

	return;
}

/* ========================================================================================================
 *
 * This is the algorithmic section of the code.
 */

/* ========================================================================================================
 *
 * Link functions.
 */

static inline void asym_link_remove(IN asym_algorithm_t *p_algo, IN asym_adata_t *p_adata, IN asym_alink_t *p_alink, IN int removal_type)
{
	osm_switch_t		*p_sw1;
	osm_switch_t		*p_sw2;
	asym_adata_t		*p_bdata;
	asym_alink_t		*p_blink;
	asym_iter_t		iter;
	asym_t			*p_asym;

	p_asym = p_algo->p_asym;

	/*
	 * If this is a real removal, then save it for calculating AR groups.
	 */
	p_alink->state = removal_type;
	if (removal_type == LINK_REMOVED) {
		alink_to_switches(p_alink, p_sw1, p_sw2);
		OSM_LOG(p_asym->p_log, OSM_LOG_VERBOSE,
			"src=%s -> dst=%s : link from %s -> %s removed\n",
			sw_name(p_algo->p_src_sw), sw_name(p_algo->p_dst_sw),
			sw_name(p_sw1), sw_name(p_sw2));

		if (!asym_set_contains(&p_algo->removals, p_alink)) {
			asym_set_insert(&p_algo->removals, p_alink);
		}
	}

	/*
	 * If the previous node is no longer connected to the SRC, then we remove all of its
	 * incoming links.
	 */
	p_bdata = alink_remote_adata(p_adata, p_alink);
	if ((p_bdata->layer == LAYER_SRC_SPINE) || (p_bdata->layer == LAYER_CORE)) {
		if (asym_node_is_not_connected(p_algo, p_bdata)) {
			foreach_incoming_link(p_bdata, p_blink, iter) {
				asym_link_remove(p_algo, p_bdata, p_blink, LINK_NA);
			}
		}
	}
}

static inline void asym_calculate_link_weights(asym_algorithm_t *p_algo, asym_adata_t *p_adata)
{
	int			outgoing_faked_count;
	int			incoming_count;
	int			outgoing_count;
	weight_t		incoming_real_weight;
	weight_t		incoming_weight;
	weight_t		average_weight;
	weight_t		link_weight;
	asym_iter_t		iter;
	asym_alink_t		*p_alink;

	outgoing_faked_count = 0;
	outgoing_count = 0;
	incoming_count = 0.0;
	incoming_weight = 0.0;
	incoming_real_weight = 0.0;
	link_weight = 0.0;

	foreach_link(p_adata, p_alink, iter) {
		if (link_is_incoming(p_adata, p_alink) && (link_is_up(p_alink) || link_is_faked(p_alink))) {
			incoming_count++;
			incoming_weight += p_alink->lw;
			incoming_real_weight += p_alink->rlw;
		} else if (link_is_outgoing(p_adata, p_alink) && (link_is_up(p_alink) || link_is_faked(p_alink))) {
			outgoing_count++;
		}

		if (link_is_outgoing(p_adata, p_alink) && link_is_faked(p_alink)) {
			outgoing_faked_count++;
		}
	}

	if (p_adata->layer == LAYER_DST_SPINE) {
		if ((outgoing_count - outgoing_faked_count) != 0) {
			link_weight = asym_min(1.0, incoming_weight / (outgoing_count - outgoing_faked_count));
		} else {
			link_weight = 0.0;
		}
	} else if (p_adata->layer == LAYER_CORE) {
		if (incoming_count == 0) {
			link_weight = 0.0;
		} else if (outgoing_count == 0) {
			link_weight = 0.0;
		} else {
			average_weight = asym_min(incoming_weight, outgoing_count) / incoming_count;
			link_weight = average_weight * asym_min(1.0, (weight_t)incoming_count / (weight_t)outgoing_count);
		}
	}

	foreach_link(p_adata, p_alink, iter) {
		if (link_is_outgoing(p_adata, p_alink)) {
			if (p_adata->layer == LAYER_SRC_SPINE) {
				p_alink->lw = incoming_real_weight;
			} else {
				p_alink->lw = link_weight;
			}
		}
	}
}

static inline void asym_recalculate_link_weights(asym_algorithm_t *p_algo, asym_adata_t *p_adata)
{
	int			outgoing_count;
	weight_t		incoming_weight;
	asym_iter_t		iter;
	asym_alink_t		*p_alink = NULL;

	incoming_weight = 0.0;
	outgoing_count = 0;

	foreach_link(p_adata, p_alink, iter) {
		if (link_is_incoming(p_adata, p_alink)) {
			incoming_weight += p_alink->lw;
		} else if (link_is_up(p_alink) && link_is_outgoing(p_adata, p_alink)) {
			outgoing_count++;
		}
	}

	/*
	 * If there are no outgoing links, the weight is zero.
	 */
	if (outgoing_count == 0) {
		if (p_alink) {
			p_alink->lw = 0.0;
		}
		return;
	}

	foreach_link(p_adata, p_alink, iter) {
		if (link_is_up(p_alink) && link_is_outgoing(p_adata, p_alink)) {
			p_alink->lw *= (weight_t)(outgoing_count+1) / (weight_t)outgoing_count;
		}
	}
}

static inline void asym_calculate_real_link_weights(asym_algorithm_t *p_algo, asym_adata_t *p_adata)
{
	asym_iter_t		iter;
	asym_alink_t		*p_alink;

	foreach_outgoing_link(p_adata, p_alink, iter) {
		p_alink->rlw = asym_min(p_adata->nw, 1.0);
	}
}

/* ========================================================================================================
 *
 * Node functions.
 */

static inline boolean_t asym_node_is_balanced(asym_algorithm_t *p_algo, asym_adata_t *p_adata)
{
	weight_t		link_weight;
	boolean_t		balanced;
	asym_iter_t		iter;
	asym_alink_t		*p_alink;

	link_weight = -1.0;
	balanced = TRUE;
	foreach_incoming_link(p_adata, p_alink, iter) {
		if (link_weight < 0.0) {
			link_weight = p_alink->lw;
		}

		if (W_gt(p_alink->lw, 0.0)) {
			balanced &= W_eq(link_weight, p_alink->lw);
		}
	}

	return balanced;
}

static inline boolean_t asym_node_is_not_balanced(asym_algorithm_t *p_algo, asym_adata_t *p_adata)
{
	return !asym_node_is_balanced(p_algo, p_adata);
}

static inline void asym_node_set_balance(asym_algorithm_t *p_algo, asym_adata_t *p_adata)
{
	p_adata->balanced = asym_node_is_balanced(p_algo, p_adata);
}

static inline void asym_calculate_node_weight(asym_algorithm_t *p_algo, asym_adata_t *p_adata)
{
	int			incoming_count;
	int			outgoing_count;
	weight_t		incoming_weight;
	asym_iter_t		iter;
	asym_alink_t		*p_alink;

	/*
	 * If we are at the source spine node, we will have only 1 link to the source.
	 */
	incoming_count = 0;
	outgoing_count = 0;
	incoming_weight = 0.0;
	foreach_link(p_adata, p_alink, iter) {
		if (link_is_up(p_alink) || link_is_faked(p_alink)) {
			if (link_is_outgoing(p_adata, p_alink)) {
				outgoing_count++;
			} else if (link_is_incoming(p_adata, p_alink)) {
				incoming_count++;
				incoming_weight += p_alink->rlw;
			}
		}
	}

	/*
	 * The SRC layer calculates the node weight differently.
	 */
	if (p_adata->layer == LAYER_SRC) {
		p_adata->nw = ceil(incoming_weight);
		return;
	}

	/*
	 * If there are no outgoing links (due to removals), then the node weight is 0.
	 */
	if (outgoing_count == 0) {
		p_adata->nw = 0.0;
		return;
	}

	/*
	 * For phase 1, all source spines have only one outgoing link to the source leaf.
	 */
	if (p_adata->layer == LAYER_SRC_SPINE) {
		outgoing_count = 1;
	}

	p_adata->outgoing_count = outgoing_count;

	if (p_algo->layer_number != LAYER_SRC_SPINE) {
		p_adata->nw = incoming_weight / outgoing_count;
	} else if (p_algo->retries == 0) {
		p_adata->nw = ceil(incoming_weight);
	} else {
		p_adata->nw = p_adata->rnw;
	}
}

static inline asym_alink_t *asym_node_balance(asym_algorithm_t *p_algo, asym_adata_t *p_adata, int removal_type)
{
	asym_alink_t		*p_alink;
	asym_alink_t		*p_clink;
	asym_iter_t		iter;

	/*
	 * Find the link with the smallest weight.
	 */
	p_clink = NULL;
	foreach_incoming_link(p_adata, p_alink, iter) {
		if (p_clink == NULL) {
			p_clink = p_alink;
		} else if (W_lt(p_alink->lw, p_clink->lw)) {
			p_clink = p_alink;
		} else if (W_eq(p_alink->lw, p_clink->lw)) {
			p_clink = link_choose(p_algo, p_adata, p_alink, p_clink);
		}
	}

	if (p_clink == NULL) {
		return NULL;
	}

	/*
	 * If we are fake removing the link, we just return it.
	 */
	if (removal_type == LINK_FAKED) {
		asym_link_remove(p_algo, p_adata, p_clink, LINK_FAKED);
		return p_clink;
	}

	/*
	 * Mark the link as removed.
	 */
	asym_link_remove(p_algo, p_adata, p_clink, LINK_REMOVED);

	/*
	 * If there are no more links towards the SRC, then we will return NULL.
	 */
	p_algo->p_last_unbalanced = p_adata;
	return asym_node_is_connected(p_algo, p_adata) ? p_clink : NULL;
}


/* ========================================================================================================
 *
 * Layer functions.
 */

static inline void asym_layer_calculate_link_weights(asym_algorithm_t *p_algo, asym_set_t *p_layer)
{
	asym_iter_t		iter;
	asym_adata_t		*p_adata;
	osm_switch_t		*p_sw;

	foreach_layer_switch(p_layer, p_sw, iter) {
		p_adata = switch_to_adata(p_sw);
		asym_calculate_link_weights(p_algo, p_adata);
	}
}

static inline void asym_layer_calculate_real_link_weights(asym_algorithm_t *p_algo, asym_set_t *p_layer)
{
	asym_iter_t		iter;
	asym_adata_t		*p_adata;
	osm_switch_t		*p_sw;

	foreach_layer_switch(p_layer, p_sw, iter) {
		p_adata = switch_to_adata(p_sw);
		asym_calculate_real_link_weights(p_algo, p_adata);
	}
}

/* -------------------------------------------------------------------------------------------------------- */

static inline void asym_layer_calculate_node_weights(asym_algorithm_t *p_algo, asym_set_t *p_layer)
{
	asym_iter_t		iter;
	asym_adata_t		*p_adata;
	osm_switch_t		*p_sw;

	foreach_layer_switch(p_layer, p_sw, iter) {
		p_adata = switch_to_adata(p_sw);
		asym_calculate_node_weight(p_algo, p_adata);
	}
}

static inline void asym_layer_set_node_balance(asym_algorithm_t *p_algo, asym_set_t *p_layer)
{
	asym_iter_t		iter;
	asym_adata_t		*p_adata;
	osm_switch_t		*p_sw;

	foreach_layer_switch(p_layer, p_sw, iter) {
		p_adata = switch_to_adata(p_sw);
		asym_node_set_balance(p_algo, p_adata);
	}
}

static inline asym_adata_t *asym_layer_find_unbalanced(asym_algorithm_t *p_algo, asym_set_t *p_layer)
{
	int			outgoing_count;
	weight_t		a_weight;
	weight_t		c_weight;
	asym_t			*p_asym;
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	asym_alink_t		*p_alink;
	asym_adata_t		*p_acand;
	asym_adata_t		*p_adata;
	osm_switch_t		*p_sw;
	boolean_t		flag;

	p_asym = p_algo->p_asym;
	p_acand = NULL;
	flag = FALSE;

	/*
	 * For non-core nodes, we score regularly.
	 */
	if (p_algo->layer_number == LAYER_DST_SPINE) {
		foreach_layer_switch(p_layer, p_sw, iter1) {
			p_adata = switch_to_adata(p_sw);
			asym_node_set_balance(p_algo, p_adata);

			if (!p_adata->balanced) {
				if (p_acand == NULL) {
					p_acand = p_adata;
				} else if (W_lt(p_adata->nw, p_acand->nw)) {
					p_acand = p_adata;
				} else if (W_eq(p_adata->nw, p_acand->nw)) {
					p_acand = node_choose(p_algo, p_adata, p_acand);
				}
			}
		}
	}

	/*
	 * For CORE nodes, we look at the node weight per link.
	 */
	if (p_algo->layer_number == LAYER_CORE) {
		foreach_layer_switch(p_layer, p_sw, iter1) {
			p_adata = switch_to_adata(p_sw);
			asym_node_set_balance(p_algo, p_adata);

			if (p_adata->balanced) {
				continue;
			}

			outgoing_count = 0;
			foreach_outgoing_link(p_adata, p_alink, iter2) {
				outgoing_count++;
			}

			if (outgoing_count == 0) {
				continue;
			}

			a_weight = p_adata->nw / outgoing_count;

			if (p_acand == NULL) {
				p_acand = p_adata;
				c_weight = a_weight;
			} else if (W_gt(a_weight, c_weight)) {
				continue;
			} else if (W_lt(a_weight, c_weight)) {
				p_acand = p_adata;
				c_weight = a_weight;
			} else if (p_adata == p_algo->p_last_unbalanced) {
				p_acand = p_adata;
				c_weight = a_weight;
				flag = TRUE;
			} else if ((p_acand != p_algo->p_last_unbalanced) && !flag) {
				p_acand = node_choose(p_algo, p_adata, p_acand);
			}
		}
	}

	OSM_LOG(p_algo->p_asym->p_log, OSM_LOG_DEBUG, "balancing %s\n", (p_acand != NULL) ? sw_name(adata_to_switch(p_acand)) : "(null)");
	return p_acand;
}

/* -------------------------------------------------------------------------------------------------------- */

static boolean_t asym_layer_1_process(asym_algorithm_t *p_algo)
{
	asym_set_t		*curr_layer;
	asym_set_t		*next_layer;
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	asym_alink_t		*p_alink;
	asym_adata_t		*p_adata;
	osm_switch_t		*p_sw;

	p_algo->layer_number = LAYER_DST;
	curr_layer = &p_algo->layers[LAYER_DST];
	next_layer = &p_algo->layers[LAYER_DST_SPINE];

	foreach_layer_switch(curr_layer, p_sw, iter1) {
		p_adata = switch_to_adata(p_sw);
		p_adata->nw = 1.0;

		foreach_outgoing_link(p_adata, p_alink, iter2) {
			if (link_is_up(p_alink)) {
				p_alink->rlw = 1.0;
				p_alink->lw = 1.0;
			}
		}
	}

	asym_layer_calculate_node_weights(p_algo, next_layer);
	return TRUE;
}

static boolean_t asym_layer_2_process(asym_algorithm_t *p_algo)
{
	asym_set_t		*curr_layer;
	asym_set_t		*next_layer;
	asym_adata_t		*p_acand;

	p_algo->layer_number = LAYER_DST_SPINE;
	curr_layer = &p_algo->layers[LAYER_DST_SPINE];
	next_layer = &p_algo->layers[LAYER_CORE];

	asym_layer_calculate_real_link_weights(p_algo, curr_layer);
	asym_layer_calculate_link_weights(p_algo, curr_layer);
	asym_layer_calculate_node_weights(p_algo, next_layer);

	layer_print_all("DST_SPINE : before balancing", p_algo, curr_layer);
	layer_print_all("CORE      : before balancing", p_algo, next_layer);

	while ((p_acand = asym_layer_find_unbalanced(p_algo, next_layer)) != NULL) {
		asym_node_balance(p_algo, p_acand, LINK_FAKED);

		asym_layer_calculate_link_weights(p_algo, curr_layer);
		asym_layer_calculate_node_weights(p_algo, curr_layer);

		asym_layer_calculate_link_weights(p_algo, next_layer);
		asym_layer_calculate_node_weights(p_algo, next_layer);
	}

	layer_print_all("DST_SPINE : after balancing", p_algo, curr_layer);
	layer_print_all("CORE      : after balancing", p_algo, next_layer);

	return TRUE;
}

static boolean_t asym_layer_3_process(asym_algorithm_t *p_algo)
{
	asym_set_t		*curr_layer;
	asym_set_t		*next_layer;
	asym_adata_t		*p_acand;
	asym_adata_t		*p_rdata;
	asym_alink_t		*p_alink;

	p_algo->layer_number = LAYER_CORE;
	curr_layer = &p_algo->layers[LAYER_CORE];
	next_layer = &p_algo->layers[LAYER_SRC_SPINE];

	asym_layer_calculate_real_link_weights(p_algo, curr_layer);
	asym_layer_calculate_link_weights(p_algo, curr_layer);
	asym_layer_calculate_node_weights(p_algo, next_layer);

	layer_print_all("CORE      : before balancing", p_algo, curr_layer);
	layer_print_all("SRC_SPINE : before balancing", p_algo, next_layer);

	p_algo->p_last_unbalanced = NULL;
	while ((p_acand = asym_layer_find_unbalanced(p_algo, next_layer)) != NULL) {
		if ((p_alink = asym_node_balance(p_algo, p_acand, LINK_REMOVED)) == NULL) {
			return FALSE;
		}

		p_rdata = alink_remote_adata(p_acand, p_alink);
		asym_calculate_node_weight(p_algo, p_rdata);
		asym_recalculate_link_weights(p_algo, p_rdata);
		asym_calculate_real_link_weights(p_algo, p_rdata);

		asym_layer_calculate_node_weights(p_algo, next_layer);
	}

	layer_print_all("CORE      : after balancing", p_algo, curr_layer);
	layer_print_all("SRC_SPINE : after balancing", p_algo, next_layer);

	return TRUE;
}

static boolean_t asym_layer_4_process(asym_algorithm_t *p_algo)
{
	int			incoming_count;
	asym_set_t		*curr_layer;
	asym_set_t		*next_layer;
	asym_iter_t		iter;
	asym_adata_t		*p_adata;
	asym_alink_t		*p_alink;
	asym_alink_t		*p_clink;

	p_algo->layer_number = LAYER_SRC_SPINE;
	curr_layer = &p_algo->layers[LAYER_SRC_SPINE];
	next_layer = &p_algo->layers[LAYER_SRC];

	p_adata = index_to_adata(p_algo->p_src_sw->re_index);

	asym_layer_calculate_real_link_weights(p_algo, curr_layer);
	asym_layer_calculate_link_weights(p_algo, curr_layer);
	asym_layer_calculate_node_weights(p_algo, next_layer);

	layer_print_all("SRC_SPINE : before balancing", p_algo, curr_layer);
	layer_print_all("SRC       : before balancing", p_algo, next_layer);

	/*
	 * All of the node weights should be limited to 1.
	 */
	incoming_count = 0;
	foreach_incoming_link(p_adata, p_alink, iter) {
		incoming_count++;
	}

	if (W_eq(ceil(p_adata->nw), (weight_t)incoming_count)) {
		return TRUE;
	}

	/*
	 * Find the link with the smallest RLW.
	 */
	p_clink = NULL;
	foreach_incoming_link(p_adata, p_alink, iter) {
		if (p_clink == NULL) {
			p_clink = p_alink;
		} else if (W_lt(p_alink->rlw, p_clink->rlw)) {
			p_clink = p_alink;
		} else if (W_eq(p_alink->rlw, p_clink->rlw)) {
			p_clink = link_choose(p_algo, p_adata, p_alink, p_clink);
		}
	}

	if (p_clink != NULL) {
		asym_link_remove(p_algo, p_adata, p_clink, LINK_REMOVED);
	}

	return (p_clink == NULL) ? TRUE : FALSE;
}

/* ========================================================================================================
 *
 * Long section.  This code handles the path from LEAF -> SPINE -> CORE -> SPINE -> LEAF.
 */

__attribute__((unused))
static void asym_algorithm_leaf_to_leaf_long(IN asym_algorithm_t *p_algo, IN osm_switch_t *p_src_sw, IN osm_switch_t *p_dst_sw)
{
	boolean_t		status;
	asym_adata_t		*p_adata;

	asym_algorithm_setup(p_algo, p_src_sw, p_dst_sw);

redo:
	asym_layer_1_process(p_algo);
	asym_layer_2_process(p_algo);

	status = asym_layer_3_process(p_algo);
	if (!status) {
		if (p_algo->retries == 0) {
			p_adata = switch_to_adata(p_src_sw);
			p_adata->rnw = p_adata->nw;
		}

		p_algo->retries++;
		asym_algorithm_reset_links(p_algo, LAYER_CORE);
		OSM_LOG(p_algo->p_asym->p_log, OSM_LOG_DEBUG, "restarting algorithm after layer 3\n");
		goto redo;
	}

	status = asym_layer_4_process(p_algo);
	if (!status) {
		if (p_algo->retries == 0) {
			p_adata = switch_to_adata(p_src_sw);
			p_adata->rnw = p_adata->nw;
		}

		p_algo->retries++;
		asym_algorithm_reset_links(p_algo, LAYER_CORE);
		OSM_LOG(p_algo->p_asym->p_log, OSM_LOG_DEBUG, "restarting algorithm after layer 4\n");
		goto redo;
	}

	return;
}

__attribute__((unused))
static void asym_algorithm_leaf_to_leaf_fc_fc_long(IN asym_algorithm_t *p_algo, IN asym_pod_t *p_src_pod, IN asym_pod_t *p_dst_pod)
{
	osm_switch_t		*p_src_sw;
	osm_switch_t		*p_dst_sw;
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	asym_node_t		*p_anode;
	asym_t			*p_asym;

	p_asym = p_algo->p_asym;

	/*
	 * Verify that there are switches which satisfy the (fully,fully) requirement.
	 */
	/*
	 * Find a pair of switches to run the reference path on.
	 */
	if ((p_src_sw = asym_find_fully_connected_switch(p_algo, p_src_pod)) == NULL) {
		OSM_LOG(p_asym->p_log, OSM_LOG_DEBUG, "FC_FC: no fully connected source switches\n");
		return;
	}

	if ((p_dst_sw = asym_find_fully_connected_switch(p_algo, p_dst_pod)) == NULL) {
		OSM_LOG(p_asym->p_log, OSM_LOG_DEBUG, "FC_FC: no fully connected destination switches\n");
		return;
	}

	/*
	 * Run the reference pair.  We run one pass through the fabric and then remember the
	 * removals.  We use these removals for the other fully connected pairs.
	 */
	OSM_LOG(p_asym->p_log, OSM_LOG_DEBUG, "FC_FC reference: %s -> %s\n", sw_name(p_src_sw), sw_name(p_dst_sw));
	asym_algorithm_leaf_to_leaf_long(p_algo, p_src_sw, p_dst_sw);

	/*
	 * Run all of the fully connected sources.
	 */
	foreach_pod_leaf(p_dst_pod, p_dst_sw, iter1) {
		p_anode = switch_to_anode(p_dst_sw);
		if (p_anode->is_fully_connected) {
			foreach_pod_leaf(p_src_pod, p_src_sw, iter2) {
				p_anode = switch_to_anode(p_src_sw);
				if (p_anode->is_fully_connected) {
					OSM_LOG(p_asym->p_log, OSM_LOG_DEBUG, "FC_FC: %s -> %s\n", sw_name(p_src_sw), sw_name(p_dst_sw));
					p_algo->p_src_sw = p_src_sw;
					p_algo->p_dst_sw = p_dst_sw;

					asym_set_remove_all(&p_algo->layers[LAYER_DST]);
					asym_set_insert(&p_algo->layers[LAYER_DST], p_dst_sw);

					asym_set_remove_all(&p_algo->layers[LAYER_SRC]);
					asym_set_insert(&p_algo->layers[LAYER_SRC], p_src_sw);

					asym_algorithm_set_ar_groups(p_algo, LAYER_DST_SPINE, LAYER_SRC);
				}
			}
		}
	}

	return;
}

__attribute__((unused))
static void asym_algorithm_leaf_to_leaf_fc_pc_long(IN asym_algorithm_t *p_algo, IN asym_pod_t *p_src_pod, IN asym_pod_t *p_dst_pod)
{
	osm_switch_t		*p_sw1;
	osm_switch_t		*p_sw2;
	osm_switch_t		*p_ref_sw;
	osm_switch_t		*p_src_sw;
	osm_switch_t		*p_dst_sw;
	asym_alink_t		*p_alink;
	asym_alink_t		*p_clink;
	asym_adata_t		*p_adata;
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	asym_iter_t		iter3;
	asym_node_t		*p_anode;
	asym_set_t		ref_removals;
	asym_t			*p_asym;

	p_asym = p_algo->p_asym;
	asym_set_create(&ref_removals);

	/*
	 * Verify that there are switches which satisfy the (fully,partial) requirement.
	 */
	if ((p_ref_sw = asym_find_fully_connected_switch(p_algo, p_src_pod)) == NULL) {
		OSM_LOG(p_asym->p_log, OSM_LOG_DEBUG, "FC_PC: no fully connected source switches\n");
		goto exit;
	}

	if ((p_dst_sw = asym_find_partially_connected_switch(p_algo, p_dst_pod)) == NULL) {
		OSM_LOG(p_asym->p_log, OSM_LOG_DEBUG, "FC_PC: no partially connected destination switches\n");
		goto exit;
	}

	/*
	 * Run all of the fully connected -> partially connected sources.
	 */
	foreach_pod_leaf(p_dst_pod, p_dst_sw, iter1) {
		p_anode = switch_to_anode(p_dst_sw);
		if (!p_anode->is_fully_connected) {

			/*
			 * Run the reference pair for this destination.  We save the removed links for later.
			 */
			OSM_LOG(p_asym->p_log, OSM_LOG_DEBUG, "FC_PC reference: %s -> %s\n", sw_name(p_ref_sw), sw_name(p_dst_sw));
			asym_algorithm_leaf_to_leaf_long(p_algo, p_ref_sw, p_dst_sw);

			asym_set_remove_all(&ref_removals);
			foreach_link_removal(&p_algo->removals, p_alink, iter3) {
				asym_set_insert(&ref_removals, p_alink);
			}

			foreach_pod_leaf(p_src_pod, p_src_sw, iter2) {
				if (p_src_sw != p_ref_sw) {
					p_anode = switch_to_anode(p_src_sw);
					if (p_anode->is_fully_connected) {
						p_adata = switch_to_adata(p_src_sw);
						p_adata->layer = LAYER_SRC;
					}
				}
			}

			/*
			 * We run through the fully connected sources just updating the removed links.
			 */
			foreach_pod_leaf(p_src_pod, p_src_sw, iter2) {
				p_anode = switch_to_anode(p_src_sw);
				if (p_anode->is_fully_connected) {
					OSM_LOG(p_asym->p_log, OSM_LOG_DEBUG, "FC_PC: %s -> %s\n", sw_name(p_src_sw), sw_name(p_dst_sw));
					p_algo->p_src_sw = p_src_sw;
					p_algo->p_dst_sw = p_dst_sw;

					asym_set_remove_all(&p_algo->layers[LAYER_DST]);
					asym_set_insert(&p_algo->layers[LAYER_DST], p_dst_sw);

					asym_set_remove_all(&p_algo->layers[LAYER_SRC]);
					asym_set_insert(&p_algo->layers[LAYER_SRC], p_src_sw);

					asym_set_remove_all(&p_algo->removals);
					foreach_link_removal(&ref_removals, p_alink, iter3) {
						alink_to_switches(p_alink, p_sw1, p_sw2);

						if ((p_sw1 != p_ref_sw) && (p_sw2 != p_ref_sw)) {
							p_clink = p_alink;
						} else if (p_sw1 == p_ref_sw) {
							p_clink = link_find(p_algo, p_src_sw, p_sw2);
						} else {
							p_clink = link_find(p_algo, p_src_sw, p_sw1);
						}

						asym_set_insert(&p_algo->removals, p_clink);
					}

					asym_algorithm_set_ar_groups(p_algo, LAYER_DST_SPINE, LAYER_SRC);
				}
			}
		}
	}

exit:
	asym_set_destroy(&ref_removals);

	return;
}

__attribute__((unused))
static void asym_algorithm_leaf_to_leaf_other_long(IN asym_algorithm_t *p_algo, IN asym_pod_t *p_src_pod, IN asym_pod_t *p_dst_pod)
{
	osm_switch_t		*p_src_sw;
	osm_switch_t		*p_dst_sw;
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	// NOTYET asym_node_t		*p_src_node;
	asym_t			*p_asym;

	p_asym = p_algo->p_asym;

	/*
	 * Verify that there are switches which satisfy the (partial,any) requirement.
	 */
#ifdef NOTYET
	if ((p_src_sw = asym_find_partially_connected_switch(p_algo, p_src_pod)) == NULL) {
		OSM_LOG(p_asym->p_log, OSM_LOG_DEBUG, "OTHER: no partially connected source switches\n");
		return;
	}
#endif

	/*
	 * Run all of the partially connected sources.
	 */
	foreach_pod_leaf(p_dst_pod, p_dst_sw, iter1) {
		foreach_pod_leaf(p_src_pod, p_src_sw, iter2) {
			// NOTYET p_src_node = switch_to_anode(p_src_sw);

			if (1) { // NOTYET !p_src_node->is_fully_connected) {
				OSM_LOG(p_asym->p_log, OSM_LOG_DEBUG, "OTHER: %s -> %s\n", sw_name(p_src_sw), sw_name(p_dst_sw));
				asym_algorithm_leaf_to_leaf_long(p_algo, p_src_sw, p_dst_sw);
				asym_algorithm_set_ar_groups(p_algo, LAYER_DST_SPINE, LAYER_SRC);
			}
		}
	}

	return;
}

/* ========================================================================================================
 *
 * Short section.  This code handles the path from LEAF -> SPINE -> LEAF.
 */
__attribute__((unused))
static void asym_algorithm_leaf_to_leaf_short(IN asym_algorithm_t *p_algo,
					      IN asym_pod_t *p_pod,
					      IN osm_switch_t *p_src_sw,
					      IN osm_switch_t *p_dst_sw)
{
	uint16_t		dst_lid;
	uint16_t		spine_lid;
	osm_switch_t		*p_sw;
	osm_switch_t		*p_spine;
	osm_ar_subgroup_t	dst_ar_group;
	asym_t			*p_asym;
	asym_set_t		spines;
	asym_iter_t		iter;

	p_asym = p_algo->p_asym;

	dst_lid = osm_node_get_base_lid(p_dst_sw->p_node, 0);
	dst_lid = cl_ntoh16(dst_lid);

	/*
	 * Setup the src leaf AR group.
	 */
	asym_set_create(&spines);
	osm_switch_clear_minhop_ports(&dst_ar_group);

	/*
	 * Any spine switch which is 1 hop away from the source and destination is in the AR group.
	 */
	foreach_pod_spine(p_pod, p_spine, iter) {
		spine_lid = osm_node_get_base_lid(p_spine->p_node, 0);
		spine_lid = cl_ntoh16(spine_lid);

		if ((osm_switch_get_least_hops(p_src_sw, spine_lid) == 1) &&
		    (osm_switch_get_least_hops(p_dst_sw, spine_lid) == 1)) {
			osm_switch_get_minhop_ports(p_src_sw, spine_lid, &dst_ar_group, NULL);
			asym_set_insert(&spines, p_spine);
		}
	}

	asym_set_remote_group_ids(p_asym, p_src_sw, p_dst_sw, &dst_ar_group);

	/*
	 * Setup the spine AR groups.
	 */
	foreach_set_switch(&spines, p_sw, iter) {
		osm_switch_clear_and_get_minhop_ports(p_sw, dst_lid, &dst_ar_group, NULL);
		asym_set_remote_group_ids(p_asym, p_sw, p_dst_sw, &dst_ar_group);
	}

	asym_set_remove_all(&spines);
	asym_set_destroy(&spines);

	return;
}

/* ========================================================================================================
 *
 * Pod section.
 */

__attribute__((unused))
static void asym_algorithm_pod_to_pod_short(IN asym_algorithm_t *p_algo, IN asym_pod_t *p_pod)
{
	asym_iter_t		iter1;
	asym_iter_t		iter2;
	osm_switch_t		*p_src_sw;
	osm_switch_t		*p_dst_sw;

	OSM_LOG(p_algo->p_asym->p_log, OSM_LOG_DEBUG, "SHORT: %d\n", p_pod->index);
	foreach_pod_leaf(p_pod, p_src_sw, iter1) {
		foreach_pod_leaf(p_pod, p_dst_sw, iter2) {
			if (p_dst_sw != p_src_sw) {
				asym_algorithm_leaf_to_leaf_short(p_algo, p_pod, p_src_sw, p_dst_sw);
			}
		}
	}
}

__attribute__((unused))
static void asym_algorithm_pod_to_pod_long(IN asym_algorithm_t *p_algo, IN asym_pod_t *p_src_pod, IN asym_pod_t *p_dst_pod)
{
	/*
	 * Run the different flavors of the algorithm.  The naming of the functions follows:
	 *	fc_fc -> fully connected to fully connected nodes
	 *	fc_pc -> fully connected to partially connected nodes
	 *	other -> partially connected to other nodes
	 */
	// NOTYET asym_algorithm_leaf_to_leaf_fc_fc_long(p_algo, p_src_pod, p_dst_pod);
	// NOTYET asym_algorithm_leaf_to_leaf_fc_pc_long(p_algo, p_src_pod, p_dst_pod);
	asym_algorithm_leaf_to_leaf_other_long(p_algo, p_src_pod, p_dst_pod);

	return;
}

int asym_algorithm_pod_to_pod(IN asym_algorithm_t *p_algo, IN asym_pod_t *p_src_pod, IN asym_pod_t *p_dst_pod)
{
	if (p_src_pod->index == p_dst_pod->index) {
		asym_algorithm_pod_to_pod_short(p_algo, p_src_pod);
	} else {
		asym_algorithm_pod_to_pod_long(p_algo, p_src_pod, p_dst_pod);
	}

	return 0;
}

